"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("util");
const url_1 = require("url"); // eslint-disable-line node/no-deprecated-api
const path = require("path");
const fs = require("fs");
const os = require("os");
const events_1 = require("events");
const pMemoize = require("p-memoize");
const filenamify = require("filenamify");
const unused_filename_1 = require("unused-filename");
const arrayUniq = require("array-uniq");
const arrayDiffer = require("array-differ");
const dateFns = require("date-fns");
const getResolutions = require("get-res");
const logSymbols = require("log-symbols");
const makeDir = require("make-dir");
const captureWebsite = require("capture-website");
const viewportList = require("viewport-list");
const template = require("lodash.template");
const plur = require("plur");
const filenamifyUrl = require("filenamify-url");
const pMap = require("p-map");
const writeFile = util_1.promisify(fs.writeFile);
const cpuCount = os.cpus().length;
const getResolutionsMemoized = pMemoize(getResolutions);
// @ts-expect-error
const viewportListMemoized = pMemoize(viewportList);
// TODO: Use private class fields when targeting Node.js 12.
class Pageres extends events_1.EventEmitter {
    constructor(options = {}) {
        var _a, _b, _c, _d;
        super();
        // Prevent false-positive `MaxListenersExceededWarning` warnings
        this.setMaxListeners(Number.POSITIVE_INFINITY);
        this.options = Object.assign({}, options);
        this.options.filename = (_a = this.options.filename) !== null && _a !== void 0 ? _a : '<%= url %>-<%= size %><%= crop %>';
        this.options.format = (_b = this.options.format) !== null && _b !== void 0 ? _b : 'png';
        this.options.incrementalName = (_c = this.options.incrementalName) !== null && _c !== void 0 ? _c : false;
        this.options.launchOptions = (_d = this.options.launchOptions) !== null && _d !== void 0 ? _d : {};
        // FIXME
        this.stats = {}; // eslint-disable-line @typescript-eslint/consistent-type-assertions
        this.items = [];
        this.sizes = [];
        this.urls = [];
        this._source = [];
        this._destination = '';
    }
    src(url, sizes, options) {
        if (url === undefined) {
            return this._source;
        }
        if (!(typeof url === 'string' && url.length > 0)) {
            throw new TypeError('URL required');
        }
        if (!(Array.isArray(sizes) && sizes.length > 0)) {
            throw new TypeError('Sizes required');
        }
        this._source.push({ url, sizes, options });
        return this;
    }
    dest(directory) {
        if (directory === undefined) {
            return this._destination;
        }
        if (!(typeof directory === 'string' && directory.length > 0)) {
            throw new TypeError('Directory required');
        }
        this._destination = directory;
        return this;
    }
    async run() {
        await Promise.all(this.src().map(async (source) => {
            const options = Object.assign(Object.assign({}, this.options), source.options);
            const sizes = arrayUniq(source.sizes.filter(size => /^\d{2,4}x\d{2,4}$/i.test(size)));
            const keywords = arrayDiffer(source.sizes, sizes);
            this.urls.push(source.url);
            if (sizes.length === 0 && keywords.includes('w3counter')) {
                return this.resolution(source.url, options);
            }
            if (keywords.length > 0) {
                return this.viewport({ url: source.url, sizes, keywords }, options);
            }
            const screenshots = await pMap(sizes, async (size) => {
                this.sizes.push(size);
                return this.create(source.url, size, options);
            }, { concurrency: cpuCount * 2 });
            this.items.push(...screenshots);
            return undefined;
        }));
        this.stats.urls = arrayUniq(this.urls).length;
        this.stats.sizes = arrayUniq(this.sizes).length;
        this.stats.screenshots = this.items.length;
        if (!this.dest()) {
            return this.items;
        }
        await this.save(this.items);
        return this.items;
    }
    successMessage() {
        const { screenshots, sizes, urls } = this.stats;
        const words = {
            screenshots: plur('screenshot', screenshots),
            sizes: plur('size', sizes),
            urls: plur('url', urls)
        };
        console.log(`\n${logSymbols.success} Generated ${screenshots} ${words.screenshots} from ${urls} ${words.urls} and ${sizes} ${words.sizes}`);
    }
    async resolution(url, options) {
        for (const item of await getResolutionsMemoized()) {
            this.sizes.push(item.item);
            this.items.push(await this.create(url, item.item, options));
        }
    }
    async viewport(viewport, options) {
        for (const item of await viewportListMemoized(viewport.keywords)) {
            this.sizes.push(item.size);
            viewport.sizes.push(item.size);
        }
        for (const size of arrayUniq(viewport.sizes)) {
            this.items.push(await this.create(viewport.url, size, options));
        }
    }
    async save(screenshots) {
        await Promise.all(screenshots.map(async (screenshot) => {
            await makeDir(this.dest());
            const dest = path.join(this.dest(), screenshot.filename);
            await writeFile(dest, screenshot);
        }));
    }
    async create(url, size, options) {
        var _a;
        const basename = path.isAbsolute(url) ? path.basename(url) : url;
        let hash = (_a = url_1.parse(url).hash) !== null && _a !== void 0 ? _a : '';
        // Strip empty hash fragments: `#` `#/` `#!/`
        if (/^#!?\/?$/.test(hash)) {
            hash = '';
        }
        const [width, height] = size.split('x');
        const filenameTemplate = template(`${options.filename}.${options.format}`);
        const now = Date.now();
        let filename = filenameTemplate({
            crop: options.crop ? '-cropped' : '',
            date: dateFns.format(now, 'yyyy-MM-dd'),
            time: dateFns.format(now, 'HH-mm-ss'),
            size,
            width,
            height,
            url: `${filenamifyUrl(basename)}${filenamify(hash)}`
        });
        if (options.incrementalName) {
            filename = unused_filename_1.default.sync(filename);
        }
        // TODO: Type this using the `capture-website` types
        const finalOptions = {
            width: Number(width),
            height: Number(height),
            delay: options.delay,
            timeout: options.timeout,
            fullPage: !options.crop,
            styles: options.css && [options.css],
            scripts: options.script && [options.script],
            cookies: options.cookies,
            element: options.selector,
            hideElements: options.hide,
            scaleFactor: options.scale === undefined ? 1 : options.scale,
            type: options.format === 'jpg' ? 'jpeg' : 'png',
            userAgent: options.userAgent,
            headers: options.headers,
            darkMode: options.darkMode,
            launchOptions: options.launchOptions
        };
        if (options.username && options.password) {
            finalOptions.authentication = {
                username: options.username,
                password: options.password
            };
        }
        const screenshot = await captureWebsite.buffer(url, finalOptions);
        screenshot.filename = filename;
        return screenshot;
    }
}
exports.default = Pageres;
// For CommonJS default export support
module.exports = Pageres;
module.exports.default = Pageres;
